/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2011, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TComPicYuv.cpp
    \brief    picture YUV buffer class
*/

#include <cstdlib>
#include <assert.h>
#include <memory.h>

#ifdef __APPLE__
#include <malloc/malloc.h>
#else
#include <malloc.h>
#endif

#include "TComPicYuv.h"

TComPicYuv::TComPicYuv()
{
#ifdef VIDYO
  m_eBcoMode        = BCO_MODE_Unsupported;
#endif

  m_apiPicBufY      = NULL;   // Buffer (including margin)
  m_apiPicBufU      = NULL;
  m_apiPicBufV      = NULL;
  
  m_piPicOrgY       = NULL;    // m_apiPicBufY + m_iMarginLuma*getStride() + m_iMarginLuma
  m_piPicOrgU       = NULL;
  m_piPicOrgV       = NULL;
  
  m_bIsBorderExtended = false;
}

TComPicYuv::~TComPicYuv()
{
}

Void TComPicYuv::create( Int iPicWidth, Int iPicHeight, UInt uiMaxCUWidth, UInt uiMaxCUHeight, UInt uiMaxCUDepth )
{
#ifdef VIDYO
  m_eBcoMode        = BCO_MODE_Rotate0Deg;
#endif

  m_iPicWidth       = iPicWidth;
  m_iPicHeight      = iPicHeight;
  
  // --> After config finished!
  m_iCuWidth        = uiMaxCUWidth;
  m_iCuHeight       = uiMaxCUHeight;
  
  m_iNumCuInWidth   = m_iPicWidth / m_iCuWidth;
  m_iNumCuInWidth  += ( m_iPicWidth % m_iCuWidth ) ? 1 : 0;
  
  m_iBaseUnitWidth  = uiMaxCUWidth  >> uiMaxCUDepth;
  m_iBaseUnitHeight = uiMaxCUHeight >> uiMaxCUDepth;
  
  m_iLumaMarginX    = g_uiMaxCUWidth  + 12; // up to 12-tap DIF
  m_iLumaMarginY    = g_uiMaxCUHeight + 12; // up to 12-tap DIF
  
  m_iChromaMarginX  = m_iLumaMarginX>>1;
  m_iChromaMarginY  = m_iLumaMarginY>>1;
  
  m_apiPicBufY      = (Pel*)xMalloc( Pel, ( m_iPicWidth       + (m_iLumaMarginX  <<1)) * ( m_iPicHeight       + (m_iLumaMarginY  <<1)));
  m_apiPicBufU      = (Pel*)xMalloc( Pel, ((m_iPicWidth >> 1) + (m_iChromaMarginX<<1)) * ((m_iPicHeight >> 1) + (m_iChromaMarginY<<1)));
  m_apiPicBufV      = (Pel*)xMalloc( Pel, ((m_iPicWidth >> 1) + (m_iChromaMarginX<<1)) * ((m_iPicHeight >> 1) + (m_iChromaMarginY<<1)));
  
  m_piPicOrgY       = m_apiPicBufY + m_iLumaMarginY   * getStride()  + m_iLumaMarginX;
  m_piPicOrgU       = m_apiPicBufU + m_iChromaMarginY * getCStride() + m_iChromaMarginX;
  m_piPicOrgV       = m_apiPicBufV + m_iChromaMarginY * getCStride() + m_iChromaMarginX;
  
  m_bIsBorderExtended = false;
  
  return;
}

Void TComPicYuv::destroy()
{
#ifdef VIDYO
  m_eBcoMode        = BCO_MODE_Unsupported;
#endif

  m_piPicOrgY       = NULL;
  m_piPicOrgU       = NULL;
  m_piPicOrgV       = NULL;
  
  if( m_apiPicBufY ){ xFree( m_apiPicBufY );    m_apiPicBufY = NULL; }
  if( m_apiPicBufU ){ xFree( m_apiPicBufU );    m_apiPicBufU = NULL; }
  if( m_apiPicBufV ){ xFree( m_apiPicBufV );    m_apiPicBufV = NULL; }
}

#ifdef VIDYO
Void TComPicYuv::createLuma( Int iPicWidth, Int iPicHeight, UInt uiMaxCUWidth, UInt uiMaxCUHeight, UInt uiMaxCUDepth, BcoMode eBcoMode )
#else
Void TComPicYuv::createLuma( Int iPicWidth, Int iPicHeight, UInt uiMaxCUWidth, UInt uiMaxCUHeight, UInt uiMaxCUDepth )
#endif
{
#ifdef VIDYO
  m_eBcoMode        = eBcoMode;
#endif

  m_iPicWidth       = iPicWidth;
  m_iPicHeight      = iPicHeight;
  
  // --> After config finished!
  m_iCuWidth        = uiMaxCUWidth;
  m_iCuHeight       = uiMaxCUHeight;
  
  m_iNumCuInWidth   = m_iPicWidth / m_iCuWidth;
  m_iNumCuInWidth  += ( m_iPicWidth % m_iCuWidth ) ? 1 : 0;
  
  m_iBaseUnitWidth  = uiMaxCUWidth  >> uiMaxCUDepth;
  m_iBaseUnitHeight = uiMaxCUHeight >> uiMaxCUDepth;
  
  m_iLumaMarginX    = g_uiMaxCUWidth  + 12; // up to 12-tap DIF
  m_iLumaMarginY    = g_uiMaxCUHeight + 12; // up to 12-tap DIF
  
  m_apiPicBufY      = (Pel*)xMalloc( Pel, ( m_iPicWidth       + (m_iLumaMarginX  <<1)) * ( m_iPicHeight       + (m_iLumaMarginY  <<1)));
  m_piPicOrgY       = m_apiPicBufY + m_iLumaMarginY   * getStride()  + m_iLumaMarginX;
  
  return;
}

Void TComPicYuv::destroyLuma()
{
  m_piPicOrgY       = NULL;
  
  if( m_apiPicBufY ){ xFree( m_apiPicBufY );    m_apiPicBufY = NULL; }
}

Pel*  TComPicYuv::getLumaAddr( int iCuAddr )
{
  Int iCuX = iCuAddr % m_iNumCuInWidth;
  Int iCuY = iCuAddr / m_iNumCuInWidth;
  
  return ( m_piPicOrgY + iCuY*m_iCuHeight*getStride() + iCuX*m_iCuWidth );
}

Pel*  TComPicYuv::getLumaAddr( Int iCuAddr, Int uiAbsZorderIdx )
{
  Int iCuX           = iCuAddr % m_iNumCuInWidth;
  Int iCuY           = iCuAddr / m_iNumCuInWidth;
  Int iOffsetCu      = iCuY*m_iCuHeight*getStride() + iCuX*m_iCuWidth;
  
  Int iCuSizeInBases = m_iCuWidth / m_iBaseUnitWidth;
  Int iRastPartIdx   = g_auiZscanToRaster[uiAbsZorderIdx];
  Int iBaseX         = iRastPartIdx % iCuSizeInBases;
  Int iBaseY         = iRastPartIdx / iCuSizeInBases;
  Int iOffsetBase    = iBaseY*m_iBaseUnitHeight*getStride() + iBaseX*m_iBaseUnitWidth;
  
  return (m_piPicOrgY + iOffsetCu + iOffsetBase);
}

Pel*  TComPicYuv::getCbAddr( int iCuAddr )
{
  Int iCuX = iCuAddr % m_iNumCuInWidth;
  Int iCuY = iCuAddr / m_iNumCuInWidth;
  
  return ( m_piPicOrgU + ( ( iCuY*m_iCuHeight*getCStride() + iCuX*m_iCuWidth )>>1 ) );
}

Pel*  TComPicYuv::getCbAddr( Int iCuAddr, Int uiAbsZorderIdx )
{
  Int iCuX           = iCuAddr % m_iNumCuInWidth;
  Int iCuY           = iCuAddr / m_iNumCuInWidth;
  Int iOffsetCu      = iCuY*m_iCuHeight*getCStride() + iCuX*m_iCuWidth;
  
  Int iCuSizeInBases = m_iCuWidth / m_iBaseUnitWidth;
  Int iRastPartIdx   = g_auiZscanToRaster[uiAbsZorderIdx];
  Int iBaseX         = iRastPartIdx % iCuSizeInBases;
  Int iBaseY         = iRastPartIdx / iCuSizeInBases;
  Int iOffsetBase    = iBaseY*m_iBaseUnitHeight*getCStride() + iBaseX*m_iBaseUnitWidth;
  
  return (m_piPicOrgU + ( ( iOffsetCu + iOffsetBase)>>1 ) );
}

Pel*  TComPicYuv::getCrAddr( int iCuAddr )
{
  Int iCuX = iCuAddr % m_iNumCuInWidth;
  Int iCuY = iCuAddr / m_iNumCuInWidth;
  
  return ( m_piPicOrgV + ( ( iCuY*m_iCuHeight*getCStride() + iCuX*m_iCuWidth )>>1 ) );
}

Pel*  TComPicYuv::getCrAddr( Int iCuAddr, Int uiAbsZorderIdx )
{
  Int iCuX           = iCuAddr % m_iNumCuInWidth;
  Int iCuY           = iCuAddr / m_iNumCuInWidth;
  Int iOffsetCu      = iCuY*m_iCuHeight*getCStride() + iCuX*m_iCuWidth;
  
  Int iCuSizeInBases = m_iCuWidth / m_iBaseUnitWidth;
  Int iRastPartIdx   = g_auiZscanToRaster[uiAbsZorderIdx];
  Int iBaseX         = iRastPartIdx % iCuSizeInBases;
  Int iBaseY         = iRastPartIdx / iCuSizeInBases;
  Int iOffsetBase    = iBaseY*m_iBaseUnitHeight*getCStride() + iBaseX*m_iBaseUnitWidth;
  
  return (m_piPicOrgV + ( ( iOffsetCu + iOffsetBase)>>1 ) );
}

Void  TComPicYuv::copyToPic (TComPicYuv*  pcPicYuvDst)
{
#ifdef VIDYO
  pcPicYuvDst->setBcoMode( m_eBcoMode );
#endif
  
  assert( m_iPicWidth  == pcPicYuvDst->getWidth()  );
  assert( m_iPicHeight == pcPicYuvDst->getHeight() );
  
  ::memcpy ( pcPicYuvDst->getBufY(), m_apiPicBufY, sizeof (Pel) * ( m_iPicWidth       + (m_iLumaMarginX   << 1)) * ( m_iPicHeight       + (m_iLumaMarginY   << 1)) );
  ::memcpy ( pcPicYuvDst->getBufU(), m_apiPicBufU, sizeof (Pel) * ((m_iPicWidth >> 1) + (m_iChromaMarginX << 1)) * ((m_iPicHeight >> 1) + (m_iChromaMarginY << 1)) );
  ::memcpy ( pcPicYuvDst->getBufV(), m_apiPicBufV, sizeof (Pel) * ((m_iPicWidth >> 1) + (m_iChromaMarginX << 1)) * ((m_iPicHeight >> 1) + (m_iChromaMarginY << 1)) );
  return;
}

Void  TComPicYuv::copyToPicLuma (TComPicYuv*  pcPicYuvDst)
{
#ifdef VIDYO
  pcPicYuvDst->setBcoMode( m_eBcoMode );
#endif

  assert( m_iPicWidth  == pcPicYuvDst->getWidth()  );
  assert( m_iPicHeight == pcPicYuvDst->getHeight() );
  
  ::memcpy ( pcPicYuvDst->getBufY(), m_apiPicBufY, sizeof (Pel) * ( m_iPicWidth       + (m_iLumaMarginX   << 1)) * ( m_iPicHeight       + (m_iLumaMarginY   << 1)) );
  return;
}

Void  TComPicYuv::copyToPicCb (TComPicYuv*  pcPicYuvDst)
{
#ifdef VIDYO
  pcPicYuvDst->setBcoMode( m_eBcoMode );
#endif

  assert( m_iPicWidth  == pcPicYuvDst->getWidth()  );
  assert( m_iPicHeight == pcPicYuvDst->getHeight() );
  
  ::memcpy ( pcPicYuvDst->getBufU(), m_apiPicBufU, sizeof (Pel) * ((m_iPicWidth >> 1) + (m_iChromaMarginX << 1)) * ((m_iPicHeight >> 1) + (m_iChromaMarginY << 1)) );
  return;
}

Void  TComPicYuv::copyToPicCr (TComPicYuv*  pcPicYuvDst)
{
#ifdef VIDYO
  pcPicYuvDst->setBcoMode( m_eBcoMode );
#endif

  assert( m_iPicWidth  == pcPicYuvDst->getWidth()  );
  assert( m_iPicHeight == pcPicYuvDst->getHeight() );
  
  ::memcpy ( pcPicYuvDst->getBufV(), m_apiPicBufV, sizeof (Pel) * ((m_iPicWidth >> 1) + (m_iChromaMarginX << 1)) * ((m_iPicHeight >> 1) + (m_iChromaMarginY << 1)) );
  return;
}


Void TComPicYuv::getLumaMinMax( Int *pMin, Int *pMax )
{
  Pel*  piY   = getLumaAddr();
  Int   iMin  = (1<<(g_uiBitDepth))-1;
  Int   iMax  = 0;
  Int   x, y;
  
  for ( y = 0; y < m_iPicHeight; y++ )
  {
    for ( x = 0; x < m_iPicWidth; x++ )
    {
      if ( piY[x] < iMin ) iMin = piY[x];
      if ( piY[x] > iMax ) iMax = piY[x];
    }
    piY += getStride();
  }
  
  *pMin = iMin;
  *pMax = iMax;
}

Void TComPicYuv::extendPicBorder ()
{
  if ( m_bIsBorderExtended ) return;
  
  xExtendPicCompBorder( getLumaAddr(), getStride(),  getWidth(),      getHeight(),      m_iLumaMarginX,   m_iLumaMarginY   );
  xExtendPicCompBorder( getCbAddr()  , getCStride(), getWidth() >> 1, getHeight() >> 1, m_iChromaMarginX, m_iChromaMarginY );
  xExtendPicCompBorder( getCrAddr()  , getCStride(), getWidth() >> 1, getHeight() >> 1, m_iChromaMarginX, m_iChromaMarginY );
  
  m_bIsBorderExtended = true;
}

Void TComPicYuv::xExtendPicCompBorder  (Pel* piTxt, Int iStride, Int iWidth, Int iHeight, Int iMarginX, Int iMarginY)
{
  Int   x, y;
  Pel*  pi;
  
  pi = piTxt;
  for ( y = 0; y < iHeight; y++)
  {
    for ( x = 0; x < iMarginX; x++ )
    {
      pi[ -iMarginX + x ] = pi[0];
      pi[    iWidth + x ] = pi[iWidth-1];
    }
    pi += iStride;
  }
  
  pi -= (iStride + iMarginX);
  for ( y = 0; y < iMarginY; y++ )
  {
    ::memcpy( pi + (y+1)*iStride, pi, sizeof(Pel)*(iWidth + (iMarginX<<1)) );
  }
  
  pi -= ((iHeight-1) * iStride);
  for ( y = 0; y < iMarginY; y++ )
  {
    ::memcpy( pi - (y+1)*iStride, pi, sizeof(Pel)*(iWidth + (iMarginX<<1)) );
  }
}


Void TComPicYuv::dump (char* pFileName, Bool bAdd)
{
  FILE* pFile;
  if (!bAdd)
  {
    pFile = fopen (pFileName, "wb");
  }
  else
  {
    pFile = fopen (pFileName, "ab");
  }
  
  Int     shift = g_uiBitIncrement;
  Int     offset = (shift>0)?(1<<(shift-1)):0;
  
  Int   x, y;
  UChar uc;
  
  Pel*  piY   = getLumaAddr();
  Pel*  piCb  = getCbAddr();
  Pel*  piCr  = getCrAddr();
  
  Pel  iMax = ((1<<(g_uiBitDepth))-1);
  
  for ( y = 0; y < m_iPicHeight; y++ )
  {
    for ( x = 0; x < m_iPicWidth; x++ )
    {
      uc = (UChar)Clip3(0, iMax, (piY[x]+offset)>>shift);
      
      fwrite( &uc, sizeof(UChar), 1, pFile );
    }
    piY += getStride();
  }
  
  for ( y = 0; y < m_iPicHeight >> 1; y++ )
  {
    for ( x = 0; x < m_iPicWidth >> 1; x++ )
    {
      uc = (UChar)Clip3(0, iMax, (piCb[x]+offset)>>shift);
      fwrite( &uc, sizeof(UChar), 1, pFile );
    }
    piCb += getCStride();
  }
  
  for ( y = 0; y < m_iPicHeight >> 1; y++ )
  {
    for ( x = 0; x < m_iPicWidth >> 1; x++ )
    {
      uc = (UChar)Clip3(0, iMax, (piCr[x]+offset)>>shift);
      fwrite( &uc, sizeof(UChar), 1, pFile );
    }
    piCr += getCStride();
  }
  
  fclose(pFile);
}

#if FIXED_ROUNDING_FRAME_MEMORY
Void TComPicYuv::xFixedRoundingPic()
{
  Int   x, y;
  Pel*  pRec    = getLumaAddr();
  Int   iStride = getStride();
  Int   iWidth  = getWidth();
  Int   iHeight = getHeight();
#if FULL_NBIT
  Int   iOffset  = ((g_uiBitDepth-8)>0)?(1<<(g_uiBitDepth-8-1)):0;
  Int   iMask   = (~0<<(g_uiBitDepth-8));
  Int   iMaxBdi = g_uiBASE_MAX<<(g_uiBitDepth-8);
#else
  Int   iOffset  = (g_uiBitIncrement>0)?(1<<(g_uiBitIncrement-1)):0;
  Int   iMask   = (~0<<g_uiBitIncrement);
  Int   iMaxBdi = g_uiBASE_MAX<<g_uiBitIncrement;
#endif

  for( y = 0; y < iHeight; y++ )
  {
    for( x = 0; x < iWidth; x++ )
    {
#if IBDI_NOCLIP_RANGE
      pRec[x] = ( pRec[x] + iOffset ) & iMask;
#else
      pRec[x] = ( pRec[x]+iOffset>iMaxBdi)? iMaxBdi : ((pRec[x]+iOffset) & iMask);
#endif
    }
    pRec += iStride;
  }

  iHeight >>= 1;
  iWidth  >>= 1;
  iStride >>= 1;
  pRec  = getCbAddr();

  for( y = 0; y < iHeight; y++ )
  {
    for( x = 0; x < iWidth; x++ )
    {
#if IBDI_NOCLIP_RANGE
      pRec[x] = ( pRec[x] + iOffset ) & iMask;
#else
      pRec[x] = ( pRec[x]+iOffset>iMaxBdi)? iMaxBdi : ((pRec[x]+iOffset) & iMask);
#endif
    }
    pRec += iStride;
  }

  pRec  = getCrAddr();

  for( y = 0; y < iHeight; y++ )
  {
    for( x = 0; x < iWidth; x++ )
    {
#if IBDI_NOCLIP_RANGE
      pRec[x] = ( pRec[x] + iOffset ) & iMask;
#else
      pRec[x] = ( pRec[x]+iOffset>iMaxBdi)? iMaxBdi : ((pRec[x]+iOffset) & iMask);
#endif
    }
    pRec += iStride;
  }
}
#endif


#ifdef VIDYO

Void TComPicYuv::setBcoMode( BcoMode eBcoMode )  
{ 
  if ( ( ( m_eBcoMode == BCO_MODE_Rotate0Deg  || m_eBcoMode == BCO_MODE_Rotate180Deg ) && ( eBcoMode == BCO_MODE_Rotate90Deg || eBcoMode == BCO_MODE_Rotate270Deg ) ) ||
       ( ( m_eBcoMode == BCO_MODE_Rotate90Deg || m_eBcoMode == BCO_MODE_Rotate270Deg ) && ( eBcoMode == BCO_MODE_Rotate0Deg  || eBcoMode == BCO_MODE_Rotate180Deg ) ) )
  {
    // Switch the width and height info

    Int iTmp;

    iTmp = m_iPicWidth;
    m_iPicWidth = m_iPicHeight;
    m_iPicHeight = iTmp;

    iTmp = m_iCuWidth;
    m_iCuWidth = m_iCuHeight;
    m_iCuHeight = iTmp;

    iTmp = m_iBaseUnitWidth;
    m_iBaseUnitWidth = m_iBaseUnitHeight;
    m_iBaseUnitHeight = iTmp;

    m_iNumCuInWidth = m_iPicWidth / m_iCuWidth;
    m_iNumCuInWidth += ( m_iPicWidth % m_iCuWidth ) ? 1 : 0;

    iTmp = m_iLumaMarginX;
    m_iLumaMarginX = m_iLumaMarginY;
    m_iLumaMarginY = iTmp;

    iTmp = m_iChromaMarginX;
    m_iChromaMarginX = m_iChromaMarginY;
    m_iChromaMarginY = iTmp;

    m_piPicOrgY = m_apiPicBufY + m_iLumaMarginY   * getStride()  + m_iLumaMarginX;
    m_piPicOrgU = m_apiPicBufU + m_iChromaMarginY * getCStride() + m_iChromaMarginX;
    m_piPicOrgV = m_apiPicBufV + m_iChromaMarginY * getCStride() + m_iChromaMarginX;
  }

  if ( m_eBcoMode != eBcoMode ) 
  {
    m_eBcoMode = eBcoMode; 
    setBorderExtension( false );
  }
}

Void TComPicYuv::rotateToPic( BcoMode eBcoMode, TComPicYuv* pcPicYuvDst, Bool bForceRotate )
{
  // Rotate the pic iff the destination YUV's BCO mode is not the same as eBcoMode
  if ( !bForceRotate && eBcoMode == pcPicYuvDst->getBcoMode() )
  {
    return;
  }

  const BcoMode eBcoModeSrc = m_eBcoMode;
  const BcoMode eBcoModeDst = eBcoMode;

  assert( eBcoModeSrc != BCO_MODE_Unsupported );
  assert( eBcoModeDst != BCO_MODE_Unsupported );
 
  BcoMode eRotate = (BcoMode) ( (eBcoModeDst - eBcoModeSrc + BCO_MODE_Unsupported) % BCO_MODE_Unsupported );

  pcPicYuvDst->setBcoMode( eBcoMode );

  Pel *ySrc = getLumaAddr();
  Pel *uSrc = getCbAddr();
  Pel *vSrc = getCrAddr();
  Int strideSrc = getStride();
  Int strideSrcC = getCStride();
  Int widthSrc = getWidth();
  Int heightSrc = getHeight();

  Pel *yDst = pcPicYuvDst->getLumaAddr();
  Pel *uDst = pcPicYuvDst->getCbAddr();
  Pel *vDst = pcPicYuvDst->getCrAddr();
  Int strideDst = pcPicYuvDst->getStride();
  Int strideDstC = pcPicYuvDst->getCStride();
  Int widthDst = pcPicYuvDst->getWidth();
  Int heightDst = pcPicYuvDst->getHeight();

  Int x, y;

  switch ( eRotate )
  {
  case BCO_MODE_Rotate0Deg:
    assert( widthSrc == widthDst );
    assert( heightSrc == heightDst );

    // Below is just a simple copy; this should not be called when no rotation is needed...
    for ( y = 0; y < heightDst; ++y, yDst += strideDst ) 
    {
      for ( x = 0; x < widthDst; ++x ) 
      {
        yDst[x] = ySrc[y*strideSrc + x];
      }
    }
    for ( y = 0; y < heightDst/2; ++y, uDst += strideDstC, vDst += strideDstC ) 
    {
      for ( x = 0; x < widthDst/2; ++x ) 
      {
        uDst[x] = uSrc[y*strideSrcC + x];
        vDst[x] = vSrc[y*strideSrcC + x];
      }
    }
    break;

  case BCO_MODE_Rotate90Deg:
    assert( widthSrc == heightDst );
    assert( heightSrc == widthDst );

    for ( y = 0; y < heightDst; ++y, yDst += strideDst ) 
    {
      for ( x = 0; x < widthDst; ++x ) 
      {
        yDst[x] = ySrc[(widthDst-x-1)*strideSrc + y];
      }
    }
    for ( y = 0; y < heightDst/2; ++y, uDst += strideDstC, vDst += strideDstC ) 
    {
      for ( x = 0; x < widthDst/2; ++x ) 
      {
        uDst[x] = uSrc[(widthDst/2-x-1)*strideSrcC + y];
        vDst[x] = vSrc[(widthDst/2-x-1)*strideSrcC + y];
      }
    }
    break;

  case BCO_MODE_Rotate180Deg:
    assert( widthSrc == widthDst );
    assert( heightSrc == heightDst );

    for ( y = 0; y < heightDst; ++y, yDst += strideDst ) 
    {
      for ( x = 0; x < widthDst; ++x ) 
      {
        yDst[x] = ySrc[(heightDst-y-1)*strideSrc + widthDst-x-1];
      }
    }
    for ( y = 0; y < heightDst/2; ++y, uDst += strideDstC, vDst += strideDstC ) 
    {
      for ( x = 0; x < widthDst/2; ++x ) 
      {
        uDst[x] = uSrc[(heightDst/2-y-1)*strideSrcC + widthDst/2-x-1];
        vDst[x] = vSrc[(heightDst/2-y-1)*strideSrcC + widthDst/2-x-1];
      }
    }
    break;

  case BCO_MODE_Rotate270Deg:
    assert( widthSrc == heightDst );
    assert( heightSrc == widthDst );
  
    for ( y = 0; y < heightDst; ++y, yDst += strideDst ) 
    {
      for ( x = 0; x < widthDst; ++x ) 
      {
        yDst[x] = ySrc[x*strideSrc + heightDst-y-1];
      }
    }
    for ( y = 0; y < heightDst/2; ++y, uDst += strideDstC, vDst += strideDstC ) 
    {
      for ( x = 0; x < widthDst/2; ++x ) 
      {
        uDst[x] = uSrc[x*strideSrcC + heightDst/2-y-1];
        vDst[x] = vSrc[x*strideSrcC + heightDst/2-y-1];
      }
    }
    break;
  }

  if ( m_bIsBorderExtended ) 
  {
    pcPicYuvDst->extendPicBorder();
  }
}

Void TComPicYuv::downSampleToPic( TComPicYuv*  pcPicYuvDst, UInt uiPadSize )
{
  // 4:1 downsample in each dimension
  assert( (m_iPicWidth>>2)  == pcPicYuvDst->getWidth()  );
  assert( (m_iPicHeight>>2) == pcPicYuvDst->getHeight() );
  
  assert( uiPadSize <= 12 );
  
  UInt x, y;

  Pel *ySrc = getLumaAddr();
  Pel *yDst = pcPicYuvDst->getLumaAddr();
  Int strideSrc = getStride();
  Int strideDst = pcPicYuvDst->getStride();
  const Int width = pcPicYuvDst->getWidth();
  const Int height = pcPicYuvDst->getHeight();

  for ( y = 0; y < height; ++y ) 
  {
    for ( x = 0; x < width; ++x )
    {
      yDst[x] = (  ySrc[              4*x]  + ySrc[              4*x+1]  + ySrc[              4*x+2]  + ySrc[              4*x+3]
                 + ySrc[strideSrc   + 4*x]  + ySrc[strideSrc   + 4*x+1]  + ySrc[strideSrc   + 4*x+2]  + ySrc[strideSrc   + 4*x+3]
                 + ySrc[strideSrc*2 + 4*x]  + ySrc[strideSrc*2 + 4*x+1]  + ySrc[strideSrc*2 + 4*x+2]  + ySrc[strideSrc*2 + 4*x+3]
                 + ySrc[strideSrc*3 + 4*x]  + ySrc[strideSrc*3 + 4*x+1]  + ySrc[strideSrc*3 + 4*x+2]  + ySrc[strideSrc*3 + 4*x+3] ) / 16;
      
    }
    ySrc += strideSrc*4;
    yDst += strideDst;
  }

  if ( uiPadSize ) 
  {
    yDst = pcPicYuvDst->getLumaAddr() - uiPadSize;
    for ( y = 0; y < height; ++y )
    {
      for ( x = 0; x < uiPadSize; ++x )
      {
        yDst[x] = yDst[uiPadSize];
        yDst[width + uiPadSize + x] = yDst[width + uiPadSize - 1];
      }
      yDst += strideDst;
    }

    yDst = pcPicYuvDst->getLumaAddr() - uiPadSize;
    for ( y = 0; y < uiPadSize; ++y ) {
      ::memcpy( yDst - (y+1) * strideDst,      yDst,                          sizeof(Pel) * strideDst );
      ::memcpy( yDst + (height+y) * strideDst, yDst + (height-1) * strideDst, sizeof(Pel) * strideDst );
    }
  }

  // TODO: Below is not really necessary as the block coding order is determined based on the luma info
  Pel *uSrc = getCbAddr();
  Pel *vSrc = getCrAddr();
  Pel *uDst = pcPicYuvDst->getCbAddr();
  Pel *vDst = pcPicYuvDst->getCrAddr();

  strideSrc = getCStride();
  strideDst = pcPicYuvDst->getCStride();

  for ( y = 0; y < height / 2; ++y ) 
  {
    for ( x = 0; x < width / 2; ++x )
    {
      uDst[x] = (  uSrc[              4*x]  + uSrc[              4*x+1]  + uSrc[              4*x+2]  + uSrc[              4*x+3]
                 + uSrc[strideSrc   + 4*x]  + uSrc[strideSrc   + 4*x+1]  + uSrc[strideSrc   + 4*x+2]  + uSrc[strideSrc   + 4*x+3]
                 + uSrc[strideSrc*2 + 4*x]  + uSrc[strideSrc*2 + 4*x+1]  + uSrc[strideSrc*2 + 4*x+2]  + uSrc[strideSrc*2 + 4*x+3]
                 + uSrc[strideSrc*3 + 4*x]  + uSrc[strideSrc*3 + 4*x+1]  + uSrc[strideSrc*3 + 4*x+2]  + uSrc[strideSrc*3 + 4*x+3] ) / 16;
      
      vDst[x] = (  vSrc[              4*x]  + vSrc[              4*x+1]  + vSrc[              4*x+2]  + vSrc[              4*x+3]
                 + vSrc[strideSrc   + 4*x]  + vSrc[strideSrc   + 4*x+1]  + vSrc[strideSrc   + 4*x+2]  + vSrc[strideSrc   + 4*x+3]
                 + vSrc[strideSrc*2 + 4*x]  + vSrc[strideSrc*2 + 4*x+1]  + vSrc[strideSrc*2 + 4*x+2]  + vSrc[strideSrc*2 + 4*x+3]
                 + vSrc[strideSrc*3 + 4*x]  + vSrc[strideSrc*3 + 4*x+1]  + vSrc[strideSrc*3 + 4*x+2]  + vSrc[strideSrc*3 + 4*x+3] ) / 16;
    }
    uSrc += strideSrc*4;
    vSrc += strideSrc*4;
    uDst += strideDst;
    vDst += strideDst;
  }
}

#endif